@@ -15,6 +15,43 @@ from ctypes import *
 logger = logging.getLogger(__name__)
 
 
+class TonlibError(Exception):
+    def __init__(self, result):
+        self.result = result
+
+    @property
+    def code(self):
+        return self.result.get('code')
+
+    def __str__(self):
+        return self.result.get('message')
+
+class LiteServerTimeout(TonlibError):
+    pass
+
+class BlockNotFound(TonlibError):
+    pass
+
+class BlockDeleted(TonlibError):
+    pass
+
+class ExternalMessageNotAccepted(TonlibError):
+    pass
+
+def parse_tonlib_error(result):
+    if result.get('@type') == 'error':
+        message = result.get('message')
+        if 'not in db' in message:
+            return BlockNotFound(result)
+        if "state already gc'd" in message:
+            return BlockDeleted(result)
+        if 'cannot apply external message to current state' in message:
+            return ExternalMessageNotAccepted(result)
+        if 'adnl query timeout' in message:
+            return LiteServerTimeout(result)
+        return TonlibError(result)
+    return None
+
 def get_tonlib_path():
     arch_name = platform.system().lower()
     machine = platform.machine().lower()
@@ -28,16 +65,6 @@ def get_tonlib_path():
         raise RuntimeError(f"Platform '{arch_name}({machine})' is not compatible yet")
     return pkg_resources.resource_filename('pytonlib', f'distlib/{arch_name}/{lib_name}')
 
-
-class TonLibWrongResult(Exception):
-    def __init__(self, description, result={}):
-        self.description = description
-        self.result = result
-
-    def __str__(self):
-        return f"{self.description} - unexpected lite server response:\n\t{json.dumps(self.result)}"
-
-
 # class TonLib for single liteserver
 class TonLib:
     def __init__(self, loop, ls_index, cdll_path=None, verbosity_level=0):
@@ -100,6 +127,9 @@ class TonLib:
             raise RuntimeError(f'Error in tonlibjson.__del__: {ee}')
 
     def send(self, query):
+        if not self._is_working:
+            raise RuntimeError(f"TonLib failed with state: {self._state}")
+
         query = json.dumps(query).encode('utf-8')
         try:
             self._tonlib_json_client_send(self._client, query)
@@ -119,6 +149,9 @@ class TonLib:
         return result
 
     def execute(self, query, timeout=10):
+        if not self._is_working:
+            raise RuntimeError(f"TonLib failed with state: {self._state}")
+
         extra_id = "%s:%s:%s" % (time.time() + timeout, self.ls_index, random.random())
         query["@extra"] = extra_id
         
@@ -170,16 +203,19 @@ class TonLib:
                     logger.critical(f"Tonlib #{self.ls_index:03d} crashed: {traceback.format_exc()}")
                     self._state = "crashed"
                 
-                if result and isinstance(result, dict) and ("@extra" in result) and (result["@extra"] in self.futures):
+                if isinstance(result, dict) and ("@extra" in result) and (result["@extra"] in self.futures):
                     try:
                         if not self.futures[result["@extra"]].done():
-                            self.futures[result["@extra"]].set_result(result)
-                            self.futures.pop(result["@extra"])
+                            tonlib_error = parse_tonlib_error(result)
+                            if tonlib_error is not None:
+                                self.futures[result["@extra"]].set_exception(tonlib_error)
+                            else:
+                                self.futures[result["@extra"]].set_result(result)
+                        self.futures.pop(result["@extra"])
                     except Exception as e:
                         logger.error(f'Tonlib #{self.ls_index:03d} receiving result exception: {e}')
         except Exception as ee:
             logger.critical(f'Task read_results failed: {ee}')
-        return
 
     async def del_expired_futures_loop(self):
         try:
@@ -187,9 +223,6 @@ class TonLib:
                 self.cancel_futures()
                 await asyncio.sleep(1)
 
-            # finished
-            await self.close()
             self.cancel_futures(cancel_all=True)
         except Exception as ee:
             logger.critical(f'Task del_expired_futures_loop failed: {ee}')
-        return
